//////////////////////////////////////////////////////////////
/////           Divers schemas de volumes finis        ///////
/////               pour pbs elliptiques 2D            ///////
/////                                                  ///////
/////              différents schémas de VF            ///////
/////                                                  ///////
/////                F. Boyer  &  S. Krell             ///////
/////                     Mars 2014                    ///////
//////////////////////////////////////////////////////////////

// Toutes ces fonctions prennent deux paramètres :
//      m        : un maillage
//      donnees  : un jeu de donnees
//

//**************************************************************************
//    Routine de calcul du coefficient de diffusion aux aretes
//        selon la méthode choisie
//**************************************************************************
function [ck]=calcul_coeff_k_interf(m,donnees);
  
  if ~isfield(donnees,"methode") then
    donnees.methode="exacte";
  end;

  select donnees.methode
   case "exacte"  // On fait le calcul exact du coefficient aux centres
                  // des aretes
    
     ck=evaluation(m.aretes(:,[_MILIEU_X,_MILIEU_Y]),donnees.coeff_k);
     
   case "arithmetique"   // On fait la moyenne arithmétique des valeurs du
                         // coefficient aux centres des mailles voisines
  
     ck=evaluation(m.centres(m.aretes(:,_K),[_X _Y]),donnees.coeff_k);
     
     temp=find(m.aretes(:,_L)>0);
     ck(temp)=0.5*(ck(temp) ...
		   + evaluation(m.centres(m.aretes(temp,_L),[_X _Y]),donnees.coeff_k));
     
   case "harmonique"   // On fait la moyenne harmonique pondérée des
                       // valeurs du coefficient aux centres des mailles voisines
 
     ck=evaluation(m.centres(m.aretes(:,_K),[_X _Y]),donnees.coeff_k);
     
     temp=find(m.aretes(:,_L)>0);
     ckL=evaluation(m.centres(m.aretes(temp,_L),[_X _Y]),donnees.coeff_k);
     
     ck(temp)=  m.aretes(temp,_DKL).*ck(temp).*ckL ...
	      ./(m.aretes(temp,_DKsigma).*ckL+m.aretes(temp,_DLsigma).*ck(temp));
  else
     disp("Vous n''avez pas choisi une methode de calcul du coeff de"+...
	  " diffusion valable!");
     halt;
  end;
endfunction



//**************************************************************************
//    Le schéma VF4 avec conditions de Dirichlet homogène
//    programmé à l'aide de boucles
//    Celui-ci n'est pas utilisé mais est seulement présenté
//    pour des raisons pédagogiques
//**************************************************************************

function [A,b]=const_schema_VF4_boucles(m,donnees)
  
  printf("Assemblage VF4 Dirichlet homogène avec des boucles\n");
  
  source=evaluation(centres,donnees.source);
  
  A=spzeros(m.nb_vol,m.nb_vol);
  b=zeros(m.nb_vol,1);
  
  
  for i=1:m.nb_are
  
    coeff_diff=donnees.coeff_k(m.aretes(i,_MILIEU_X),m.aretes(i,_MILIEU_Y));

    tauKL = coeff_diff*m.aretes(i,_MES)/m.aretes(i,_DKL);
    
    A(m.aretes(i,_K),m.aretes(i,_K)) = A(m.aretes(i,_K),m.aretes(i,_K)) + tauKL;
    
    b(m.aretes(i,_K)) = b(m.aretes(i,_K)) ...
	+ (m.aretes(i,_MES_K_DEB)+m.aretes(i,_MES_K_FIN))...
	   * donnees(m.aretes(i,_K));
    
    
    if (m.aretes(i,_L)>0) then
      A(m.aretes(i,_K),m.aretes(i,_L)) = - tauKL;
      A(m.aretes(i,_L),m.aretes(i,_L)) = A(m.aretes(i,_L),m.aretes(i,_L)) + tauKL;
      A(m.aretes(i,_L),m.aretes(i,_K)) = - tauKL;

      b(m.aretes(i,_L)) = b(m.aretes(i,_L)) ...
	+ (m.aretes(i,_MES_L_DEB)+m.aretes(i,_MES_L_FIN))...
	   * source(m.aretes(i,_L));      
    end;

  end;

endfunction


//**************************************************************************
//    Le schéma VF4 
//    programmé à l'aide d'opérations vectorielles pour des raisons d'efficacité 
//    Prise en compte des CL de Dirichlet non-homogène (_LABEL=-1)
//                    des CL de Neumann homogène  (_LABEL<-1)
//**************************************************************************

function [A,b]=const_schema_VF4(m,donnees)

  printf("Assemblage VF4\n");
  
  // On évalue le terme source aux centres des mailles
  
  source=evaluation(m.centres,donnees.source);
  
  // On calcule le coefficient de diffusion (scalaire) sur chaque arête
  // Ce calcul depend du choix de donnees.methode
  
  coeff_diff=calcul_coeff_k_interf(m,donnees);
  
  indiA=[];
  indjA=[];
  valA=[];
  
  indib=[];
  indjb=[];
  valb=[];
  
  // Calcul de la transmissivite de chaque arête 
  
  tauKL=coeff_diff(:).*m.aretes(:,_MES)./m.aretes(:,_DKL);
  
  /// Les aretes du bord

  /// Recherche des arêtes pour la condition de Dirichlet non homogène
  temp=find(m.aretes(:,_LABEL)==-1);

  // Evaluation des données au bord
  ubord=evaluation(...
      0.5*(m.sommets(m.aretes(temp,_DEB),[_X _Y])+m.sommets(m.aretes(temp,_FIN),[_X _Y]))...
      ,donnees.bordD);

  indiA=m.aretes(temp,_K);
  indjA=indiA;
  valA=tauKL(temp);                     

  indib=[indib; m.aretes(temp,_K)];
  valb=[valb; tauKL(temp).*ubord];


  /// Toutes les autres arêtes du bord sont supposées 
  /// correspondre à des CL de Neumann non homogène
  /// Donc elles ne contribuent pas au schéma
  
  
  /// La partie du terme source qui concerne tout le monde
  
  indib=[indib; m.aretes(:,_K)];
  valb=[valb; (m.aretes(:,_MES_K_DEB)+m.aretes(:,_MES_K_FIN)).*source(m.aretes(:,_K))];

  /// On traite maintenant les aretes interieures
  
  temp=find(m.aretes(:,_L)>0);

  indiA=[indiA; m.aretes(temp,_K)];
  indjA=[indjA; m.aretes(temp,_K)];
  valA=[valA; tauKL(temp)];  
  
  indiA=[indiA; m.aretes(temp,_K)];
  indjA=[indjA; m.aretes(temp,_L)];
  valA=[valA; -tauKL(temp)];
  
  indiA=[indiA; m.aretes(temp,_L)];
  indjA=[indjA; m.aretes(temp,_K)];
  valA=[valA; -tauKL(temp)];  

  indiA=[indiA; m.aretes(temp,_L)];
  indjA=[indjA; m.aretes(temp,_L)];
  valA=[valA; tauKL(temp)];  
  
  indib=[indib; m.aretes(temp,_L)];
  valb=[valb; (m.aretes(temp,_MES_L_DEB)+m.aretes(temp,_MES_L_FIN)).*source(m.aretes(temp,_L))];


  /// Construction proprement dite
  
  indjb=ones(indib);  
  A=sparse([indiA indjA], valA);
  b=sparse([indib indjb], valb);
  b=full(b);
  
endfunction


//**************************************************************************
//    Le schéma DDFV pour une diffusion scalaire
//    programmé à l'aide d'opérations vectorielles
//    Prise en compte des CL de Dirichlet non-homogène (_LABEL=-1)
//                    des CL de Neumann homogène  (_LABEL<-1)
//**************************************************************************


function [A,b]=const_schema_DDFV_iso(m,donnees)

  printf("Assemblage DDFV isotrope\n");
  
   printf('La fonction ''const_schema_DDFV_iso'' n''est pas encore programmee !\n');
   abort
   
   
   
   //////////////////////////////////////////////////////////////////////
   //     #              #######     #      #####   ######   ####### 
   //     #              #           #        #     #     #  #       
   //    ###             #          ###       #     #     #  #       
   //    # #             #####      # #       #     ######   #####   
   //   #####            #         #####      #     #   #    #       
   //   #   #            #         #   #      #     #    #   #       
   //  ##   ##           #        ##   ##   #####   #     #  ####### 
   //////////////////////////////////////////////////////////////////////
   
endfunction


//**************************************************************************
//    Le schéma DDFV pour une diffusion matricielle (non isotrope)
//    programmé à l'aide d'opérations vectorielles
//    Prise en compte des CL de Dirichlet non-homogène (_LABEL=-1)
//                    des CL de Neumann homogène  (_LABEL<-1)
//**************************************************************************

function [A,b]=const_schema_DDFV_aniso(m,donnees)

  printf("Assemblage DDFV Anisotrope\n");

   printf('La fonction ''const_schema_DDFV_aniso'' n''est pas encore programmee !\n');
   abort
   
   
   
   //////////////////////////////////////////////////////////////////////
   //     #              #######     #      #####   ######   ####### 
   //     #              #           #        #     #     #  #       
   //    ###             #          ###       #     #     #  #       
   //    # #             #####      # #       #     ######   #####   
   //   #####            #         #####      #     #   #    #       
   //   #   #            #         #   #      #     #    #   #       
   //  ##   ##           #        ##   ##   #####   #     #  ####### 
   //////////////////////////////////////////////////////////////////////
   
endfunction

//**************************************************************************
//    Routine de calcul du tenseur de diffusion sur un diamant
//    selon la méthode choisie
//**************************************************************************
function [ADNN,ADNNS,ADNSNS]=calcul_coeff_mat(m,donnees);

  if ~isfield(donnees,'methode') then
    donnees.methode='exacte';
  end;
  
  
  centres_aretes=m.aretes(:,[_MILIEU_X,_MILIEU_Y]);
  
  ADxx=evaluation(centres_aretes,donnees.coeff_mat.Axx);
  ADxy=evaluation(centres_aretes,donnees.coeff_mat.Axy);
  ADyy=evaluation(centres_aretes,donnees.coeff_mat.Ayy);
  
  ADN_X=ADxx.*m.aretes(:,_N_KL_X)+ADxy.*m.aretes(:,_N_KL_Y);
  ADN_Y=ADxy.*m.aretes(:,_N_KL_X)+ADyy.*m.aretes(:,_N_KL_Y);
  
  ADNS_X=ADxx.*m.aretes(:,_NS_DEBFIN_X)+ADxy.*m.aretes(:,_NS_DEBFIN_Y);
  ADNS_Y=ADxy.*m.aretes(:,_NS_DEBFIN_X)+ADyy.*m.aretes(:,_NS_DEBFIN_Y);
  
  ADNN=ADN_X.*m.aretes(:,_N_KL_X)+ADN_Y.*m.aretes(:,_N_KL_Y);
  ADNNS=ADN_X.*m.aretes(:,_NS_DEBFIN_X)+ADN_Y.*m.aretes(:,_NS_DEBFIN_Y);
  ADNSNS=ADNS_X.*m.aretes(:,_NS_DEBFIN_X)+ADNS_Y.*m.aretes(:,_NS_DEBFIN_Y);
  
  AKxx=evaluation(m.centres,donnees.coeff_mat.Axx);
  AKxy=evaluation(m.centres,donnees.coeff_mat.Axy);
  AKyy=evaluation(m.centres,donnees.coeff_mat.Ayy);
  
  
  for i=1:m.nb_are
    if (m.aretes(i,_L)>0) then
      
      nbK=m.aretes(i,_K);
      nbL=m.aretes(i,_L);
      
      AKN_X=AKxx(nbK).*m.aretes(i,_N_KL_X)+AKxy(nbK).*m.aretes(i,_N_KL_Y);
      AKN_Y=AKxy(nbK).*m.aretes(i,_N_KL_X)+AKyy(nbK).*m.aretes(i,_N_KL_Y);
      
      AKNS_X=AKxx(nbK).*m.aretes(i,_NS_DEBFIN_X)+AKxy(nbK).*m.aretes(i,_NS_DEBFIN_Y);
      AKNS_Y=AKxy(nbK).*m.aretes(i,_NS_DEBFIN_X)+AKyy(nbK).*m.aretes(i,_NS_DEBFIN_Y);
      
      AKNN=AKN_X.*m.aretes(i,_N_KL_X)+AKN_Y.*m.aretes(i,_N_KL_Y);
      AKNNS=AKN_X.*m.aretes(i,_NS_DEBFIN_X)+AKN_Y.*m.aretes(i,_NS_DEBFIN_Y);
      AKNSNS=AKNS_X.*m.aretes(i,_NS_DEBFIN_X)+AKNS_Y.*m.aretes(i,_NS_DEBFIN_Y);
      
      ALN_X=AKxx(nbL).*m.aretes(i,_N_KL_X)+AKxy(nbL).*m.aretes(i,_N_KL_Y);
      ALN_Y=AKxy(nbL).*m.aretes(i,_N_KL_X)+AKyy(nbL).*m.aretes(i,_N_KL_Y);
      
      ALNS_X=AKxx(nbL).*m.aretes(i,_NS_DEBFIN_X)+AKxy(nbL).*m.aretes(i,_NS_DEBFIN_Y);
      ALNS_Y=AKxy(nbL).*m.aretes(i,_NS_DEBFIN_X)+AKyy(nbL).*m.aretes(i,_NS_DEBFIN_Y);
      
      ALNN=ALN_X.*m.aretes(i,_N_KL_X)+ALN_Y.*m.aretes(i,_N_KL_Y);
      ALNNS=ALN_X.*m.aretes(i,_NS_DEBFIN_X)+ALN_Y.*m.aretes(i,_NS_DEBFIN_Y);
      ALNSNS=ALNS_X.*m.aretes(i,_NS_DEBFIN_X)+ALNS_Y.*m.aretes(i,_NS_DEBFIN_Y);
      
      select donnees.methode
        
      case "arithmetique"   // On fait la moyenne arithmétique des valeurs du
                            // coefficient aux centres des mailles voisines
        
        ADNN(i)=0.5*(AKNN+ALNN);
        
        ADNNS(i)=0.5*(AKNNS+ALNNS);
        
        ADNSNS(i)=0.5*(AKNSNS+ALNSNS);
	
      case "harmonique"   // On fait la moyenne harmonique pondérée des
                          // valeurs du coefficient aux centres des mailles voisines
                          // selon la méthode m-DDFV

        ADNN(i)=((m.aretes(i,_DKsigma)+m.aretes(i,_DLsigma))*AKNN*ALNN) ...
                /(m.aretes(i,_DLsigma)*AKNN+m.aretes(i,_DKsigma)*ALNN);
        
        ADNNS(i)=(m.aretes(i,_DLsigma)*ALNNS*AKNN+ ...
                 +m.aretes(i,_DKsigma)*AKNNS*ALNN) ...
                /(m.aretes(i,_DLsigma)*AKNN+m.aretes(i,_DKsigma)*ALNN);
        
        ADNSNS(i)=(m.aretes(i,_DLsigma)*ALNSNS+m.aretes(i,_DKsigma)*AKNSNS) ...
                      /(m.aretes(i,_DKsigma)+m.aretes(i,_DLsigma)) ...
                  -m.aretes(i,_DLsigma)*m.aretes(i,_DKsigma)...
		     /(m.aretes(i,_DKsigma)+m.aretes(i,_DLsigma)) ...
                     *(AKNNS-ALNNS)^2/(m.aretes(i,_DLsigma)*AKNN+m.aretes(i,_DKsigma)*ALNN);
        
      end
    end
  end

endfunction

//**************************************************************************
//    Calcul de la fonction de courant associée à la solution
//    du problème elliptique calculé par VF4 sur un maillage
//    orthogonal admissible
//    Ne fonctionne pas avec DDFV !!!
//**************************************************************************

function [courant]=calcul_courant(m,donnees,u)
   
   coeff_diff=evaluation(m.aretes(:,[_MILIEU_X,_MILIEU_Y],donnees.coeff_k);
   flux=zeros(m.nb_are,1);

   temp=find(m.aretes(:,_LABEL)>=0); // Aretes interieures
   flux(temp)=(u(m.aretes(temp,_L))-u(m.aretes(temp,_K)))./m.aretes(temp,_DKL).*m.aretes(temp,_MES);

   if isfield(donnees,'beta') then
      temp=find(m.aretes(:,_LABEL)==1); // Aretes correspondant à
                                        // d'éventuelles fractures
      flux(temp)=flux(temp)*donnees.beta.*m.aretes(temp,_DKL)./(1+ ...
						  donnees.beta*m.aretes(temp,_DKL));
   end;

   temp=find(m.aretes(:,_LABEL)==-1); // aretes du bord Dirichlet
   ubord=evaluation(m.sommets(m.aretes(temp,_DEB),[_X _Y]),donnees.bordD)
   flux(temp)=(ubord-u(m.aretes(temp,_K)))./m.aretes(temp,_DKL).*m.aretes(temp,_MES);
   


   // Pour eviter des problemes d'orientation

   signe= (m.sommets(m.aretes(:,_DEB),_X)-m.sommets(m.aretes(:,_FIN),_X)).*m.aretes(:,_N_KL_Y) ...
         -(m.sommets(m.aretes(:,_DEB),_Y)-m.sommets(m.aretes(:,_FIN),_Y)).*m.aretes(:,_N_KL_X);
   signe=signe./abs(signe);

   traite=zeros(m.nb_som,1);
   courant=zeros(m.nb_som,1);

   traite(1)=1;
   totaltraite=0;
   
   while (totaltraite<m.nb_som)
     for i=1:m.nb_are
      if ((traite(m.aretes(i,_DEB))==1) & (traite(m.aretes(i,_FIN))==0)) then
          traite(m.aretes(i,_FIN))=1;
          courant(m.aretes(i,_FIN))=courant(m.aretes(i,_DEB))+signe(i)*flux(i);
      end;

      if ((traite(m.aretes(i,_DEB))==0) & (traite(m.aretes(i,_FIN))==1)) then
          traite(m.aretes(i,_DEB))=1;
          courant(m.aretes(i,_DEB))=courant(m.aretes(i,_FIN))-signe(i)*flux(i);
      end;    
     end;

    totaltraite=sum(traite);
  end;

endfunction
 
//**************************************************************************
//    Routine d'interpolation de la fonction de courant (définie sur les sommets)
//    sur une grille cartésienne.
//     
//    Ceci est nécessaire pour tracer les lignes de courant dans Scilab
//
//    Ne marche que sur les triangles
//    et si le domaine est contenu dans le carré unité
//**************************************************************************  

function [phi]=interpole(m,courant,Nx,Ny);

   max_courant=max(courant);
   min_courant=min(courant);
  
   // On cherche les triangles
   
   tri=zeros(m.nb_vol,7);
   tri(:,1)=1;
   for i=1:m.nb_are
      j=m.aretes(i,_K);
      tri(j,tri(j,1)+1)=m.aretes(i,_DEB);
      tri(j,tri(j,1)+2)=m.aretes(i,_FIN);
      tri(j,1)=tri(j,1)+2;
      
      j=m.aretes(i,_L);
      if (j>0) then
	tri(j,tri(j,1)+1)=m.aretes(i,_DEB);
	tri(j,tri(j,1)+2)=m.aretes(i,_FIN);
	tri(j,1)=tri(j,1)+2;
      end;
   end;


   tri=tri(:,2:7)

   tri=gsort(tri,'c','i');
   tri=tri(:,[1 3 5]);


   x=linspace(0,1,Nx);
   y=linspace(0,1,Ny);
   [X Y] = meshgrid(x,y);
   phi=zeros(Nx,Ny);
   for i=1:size(X,1)
      for j=1:size(X,2)
	ptx=X(i,j);
	pty=Y(i,j);
	coord2= ( (ptx-m.sommets(tri(:,1),_X)).*(m.sommets(tri(:,3),_Y)-m.sommets(tri(:,1),_Y)) ...
                 -(pty-m.sommets(tri(:,1),_Y)).*(m.sommets(tri(:,3),_X)-m.sommets(tri(:,1),_X)))...
              ./(  (m.sommets(tri(:,2),_X)-m.sommets(tri(:,1),_X)).*(m.sommets(tri(:,3),_Y)-m.sommets(tri(:,1),_Y)) ...
                 -(m.sommets(tri(:,2),_Y)-m.sommets(tri(:,1),_Y)).*(m.sommets(tri(:,3),_X)-m.sommets(tri(:,1),_X)));
	coord3= ( (ptx-m.sommets(tri(:,1),_X)).*(m.sommets(tri(:,2),_Y)-m.sommets(tri(:,1),_Y)) ...
                 -(pty-m.sommets(tri(:,1),_Y)).*(m.sommets(tri(:,2),_X)-m.sommets(tri(:,1),_X)))...
              ./(  (m.sommets(tri(:,3),_X)-m.sommets(tri(:,1),_X)).*(m.sommets(tri(:,2),_Y)-m.sommets(tri(:,1),_Y)) ...
                 -(m.sommets(tri(:,3),_Y)-m.sommets(tri(:,1),_Y)).*(m.sommets(tri(:,2),_X)-m.sommets(tri(:,1),_X)));
	coord1=1-coord2-coord3;

	[temp]=find(abs(coord1)+abs(coord2)+abs(coord3)-1<10^(-10),1);

	if (length(temp)==0) then
	   phi(i,j)=max_courant+100*(max_courant-min_courant);
	else
	phi(i,j)=coord1(temp)*courant(tri(temp,1))+coord2(temp)*courant(tri(temp,2)) ...
                +coord3(temp)*courant(tri(temp,3));
	end;
      end;
   end;
   
   phi=phi';

endfunction